#include "myS3.h"
#include "UtilityManager.h"
#include <curl/curl.h>
#include <openssl/evp.h>
#include <sys/socket.h>       /* for AF_INET */
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <openssl/hmac.h>
#include "simplexml.h"


using namespace MyS3MultiPartUploader;

char* MyS3::HttpVerb[3] = {"GET", "PUT", "POST"};
bool MyS3::isKeyOK = false;
char MyS3::accessKey[32];
char MyS3::secretAccessKey[64];
char MyS3::keyPath[256];

int mySimpleXmlCallback(const char *elementPath, const char *data, int dataLen, void *callbackData);
size_t myCurlCallback(void *ptr, size_t size, size_t nmemb, void *user);
size_t writeHeader( void *ptr, size_t size, size_t nmemb, void *user);
int myDebugFunc(CURL * pCurl, curl_infotype, char* ptrData , size_t nDataSize, void * userData);

MyS3::MyS3(char* newKeyPath)
{
	if (newKeyPath == NULL)
	{
		newKeyPath = "/home/nick/.ec2/s3key.txt";
	}
	initializeKey(newKeyPath);
}

void MyS3::initializeKey(const char* newKeyPath)
{
	if (!isKeyOK)
	{
		strcpy(keyPath, newKeyPath);
		isKeyOK = readKeyFile();
	}
}

bool MyS3::readKeyFile()
{
	FILE* fd = fopen(keyPath, "r");
	if (fd)
	{
		fgets(accessKey, 32, fd);
		accessKey[strlen(accessKey) - 1]= '\0';
		fgets(secretAccessKey, 64, fd);
		secretAccessKey[strlen(secretAccessKey) - 1] = '\0';
		fclose(fd);
		return true;
	}
	return false;
}


bool MyS3::multiPartUpload(const string& bucketName, const string& objectName, const string& fileName, size_t nSuggestedPartNumber)
{
	if (initialize(bucketName, objectName, fileName, nSuggestedPartNumber))
	{
		if (initMultiPartUpload())
		{
			return true;
		}
	}
	return false;
}

bool MyS3::initialize(const string& bucketName, const string& objectName, const string& fileName, size_t nSuggestedPartNumber)
{
	multiPartDesc.bucket = bucketName;
	multiPartDesc.object = objectName;
	FILE* stream = fopen(fileName.c_str(), "rb");
	if (stream != NULL)
	{
		fseek(stream, 0L, SEEK_END);
		long fileSize = ftell(stream);
		long partSize = fileSize / nSuggestedPartNumber;
		long lastPartSize = 0;
		// check part size if bigger than minimum of 5M requirement
		#define MinimumPartSize 5*1024*1024
		if ( partSize < MinimumPartSize)
		{
			nSuggestedPartNumber = fileSize / MinimumPartSize;
		}
		partSize = fileSize / nSuggestedPartNumber;
		lastPartSize = fileSize % nSuggestedPartNumber;

		multiPartDesc.partCount = nSuggestedPartNumber;
		multiPartDesc.partDescVector.clear();
		long offset = 0;
		int i;
		PartDesc partDesc;
		for (i = 0; i < nSuggestedPartNumber; i ++)
		{
			partDesc.partId = i + 1;
			partDesc.offset = offset;
			partDesc.size = partSize;
			offset += partSize;
			multiPartDesc.partDescVector.push_back(partDesc);
		}

		if (lastPartSize != 0)
		{
			partDesc.partId = i + 1;
			partDesc.offset = offset;
			partDesc.size = lastPartSize;
			multiPartDesc.partDescVector.push_back(partDesc);
			multiPartDesc.partCount ++;
		}
		return true;
	}
	return false;
}

bool MyS3::initMultiPartUpload()
{
	string str;
	StringMap stringMap;
	string strToSign, strSignature, strPostRequest, strUrl, strSubUrl, strCustomRequest;
	string strContentType = "application/vnd.rn-realmedia-vbr";
	string strTimeStamp = getTimestamp();
	strPostRequest = "/" + multiPartDesc.object + "?uploads";

	FILE_LOG(logDEBUG)<<"[strPostRequest]"<< strPostRequest;
	strSubUrl = "/" + multiPartDesc.bucket + strPostRequest;

	strUrl = "http://s3.amazonaws.com/" + multiPartDesc.bucket ;
	FILE_LOG(logDEBUG)<<"[strUrl]"<< strUrl;
	strToSign = "POST\n\n";
	strToSign += strContentType;
	strToSign += "\n";
	strToSign += strTimeStamp;
	strToSign += "\n";
	strToSign += strSubUrl;

	FILE_LOG(logDEBUG)<<"[strToSign]"<< strToSign;

	string strKey = secretAccessKey;
	//MySH1 mySH1(string(secretAccessKey));
	MySH1 mySH1(strKey);
	mySH1.update(strToSign);
	string strSH1 = mySH1.finalize();

	MyBase64Encoder myBase64Encoder;

	myBase64Encoder.update(strSH1);
	strSignature = myBase64Encoder.finalize();
	FILE_LOG(logDEBUG)<<"[strSignature]"<< strSignature;
	stringMap["UploadId"] = "";
	SimpleXml simple;
	simplexml_initialize(&simple, mySimpleXmlCallback, &stringMap);

	CURL* pCurl = curl_easy_init();

	//curl_easy_setopt(pCurl, CURLOPT_POST, 1L);

	//curl_easy_setopt(pCurl, CURLOPT_POSTFIELDSIZE, strPostRequest.size());
	//curl_easy_setopt(pCurl, CURLOPT_COPYPOSTFIELDS, strPostRequest.c_str());
	strCustomRequest = "POST " + strPostRequest;
	curl_easy_setopt(pCurl, CURLOPT_CUSTOMREQUEST, strCustomRequest.c_str());
	//curl_easy_setopt(pCurl, CURLOPT_POSTFIELDS, NULL);
	//FILE_LOG(logDEBUG)<<"[strPostRequest]"<< strPostRequest;
	//

	curl_easy_setopt(pCurl, CURLOPT_VERBOSE, 1L);
	curl_easy_setopt(pCurl, CURLOPT_WRITEDATA, &simple);
	curl_easy_setopt(pCurl, CURLOPT_WRITEFUNCTION, myCurlCallback);
	curl_easy_setopt(pCurl, CURLOPT_HEADERFUNCTION, writeHeader);
	curl_easy_setopt(pCurl, CURLOPT_WRITEHEADER, &resultSet);

	curl_easy_setopt(pCurl, CURLOPT_DEBUGFUNCTION, myDebugFunc);

	//curl_easy_setopt(pCurl, CURLOPT_HEADER, 1L);
	//curl_easy_setopt(pCurl, CURLOPT_NOBODY, 1L);


	///////////////////////////////////////
	// see if it works
	curl_easy_setopt(pCurl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
	struct curl_slist *slist=NULL;
	str = "Host: " + multiPartDesc.bucket  + ".s3.amazonaws.com";
	//str = "Host: s3.amazonaws.com";
	slist = curl_slist_append(slist, str.c_str());

	str = "Date: ";
	str += strTimeStamp;
	FILE_LOG(logDEBUG)<<"[Date]"<< str;
	slist = curl_slist_append(slist, str.c_str());

	str = "Authorization: AWS ";
	str += accessKey;
	str += ":";
	str += strSignature;
	FILE_LOG(logDEBUG)<<"[Authorization]"<< str;
	slist = curl_slist_append(slist, str.c_str());

	slist = curl_slist_append(slist, "Accept-Encoding: identity");
	slist = curl_slist_append(slist, "Transfer-Encoding:");
	str = "Content-Type: " + strContentType;
	slist = curl_slist_append(slist, str.c_str());
	//slist = curl_slist_append(slist, "Accept:");
	slist = curl_slist_append(slist, "User-Agent: Mozilla/4.0 (Compatible; ubuntu32; MyS3 1.0; Linux)");
	slist = curl_slist_append(slist, "Connection: Close" );
	if (slist != NULL)
	{
		curl_easy_setopt(pCurl, CURLOPT_HTTPHEADER, slist);
	}
	curl_easy_setopt(pCurl, CURLOPT_URL, strUrl.c_str());

	curl_easy_perform(pCurl);

	curl_easy_cleanup(pCurl);
	curl_slist_free_all(slist); /* free the list again */
	simplexml_deinitialize(&simple);

	multiPartDesc.uploadId = stringMap["UploadId"];
	FILE_LOG(logDEBUG)<<"[UploadId]"<< multiPartDesc.uploadId;

}


std::string MyS3::getTimestamp()
{
	time_t after = time(NULL);
	char buffer[32];
	strftime(buffer, 32, "%a, %d %b %Y %H:%M:%S +0000", gmtime(&after));
	std::string in = buffer;
	//return urlEncode(in);
	return in;
}


int myDebugFunc(CURL * pCurl, curl_infotype infoType, char* ptrData , size_t nDataSize, void * userData)
{
	static FILE* streamDebug = NULL;
	if (streamDebug == NULL)
	{
		streamDebug = fopen("curlDebug.log", "w");
	}

	switch (infoType)
	{
	case CURLINFO_TEXT:
		fprintf(streamDebug, "%s", "[CURLINFO_TEXT]:\n");
		break;
	case CURLINFO_HEADER_IN:
		fprintf(streamDebug, "%s", "[CURLINFO_HEADER_IN]:\n");

		break;
	case CURLINFO_HEADER_OUT:
		fprintf(streamDebug, "%s", "[CURLINFO_HEADER_OUT]:\n");
		break;
	case CURLINFO_DATA_IN:
		fprintf(streamDebug, "%s", "[CURLINFO_DATA_IN]:\n");
		break;
	case CURLINFO_DATA_OUT:
		fprintf(streamDebug, "%s", "[CURLINFO_DATA_OUT]:\n");
		break;
	}
	fwrite(ptrData, 1, nDataSize, streamDebug);
	fprintf(streamDebug, "%s", "\n");
	fflush(streamDebug);
	return 0;
}

size_t writeHeader(void *ptr, size_t size, size_t nmemb, void *user)
{
	char header[32];
	char body[256];
	int len;
	StringPairVector* my= (StringPairVector*)user;
	//printf("size = %d, nmemb=%d\n", size, nmemb);
	char* str = (char*)ptr;
	char* sub = strstr(str, ": ");
	if (sub == NULL)
	{
		// actually the first line is NOT a header, i.e. "http 403 forbidden"
		len = size * nmemb - 2;
		if (len > 0 )
		{
			memcpy(body, ptr, len);
			body[len] = '\0';
			my->push_back(StringStringPair(std::string("status"), std::string(body)));
		}
	}
	else
	{
		len = sub - str;
		memcpy(header, str, len);
		header[len] = '\0';
		// remove the trailing \r\n
		len = size * nmemb - len - 2 - 2;
		memcpy(body, sub + 2, len);
		body[len] = '\0';
		my->push_back(StringStringPair(std::string(header), std::string(body)));
	}
	FILE_LOG(logDEBUG)<<"[writeHeader]"<<str;
	return size * nmemb;
}

size_t myCurlCallback(void *ptr, size_t size, size_t nmemb, void *user)
{
	SimpleXml*simple = (SimpleXml*)user;
	char* data = (char*) ptr;
	size_t result = size*nmemb;
	FILE_LOG(logDEBUG)<<"[myCurlCallback]"<<data;
	if (simplexml_add(simple, data, result))
	{
		return 0;
	}
	else
	{
		return result;
	}
}

int mySimpleXmlCallback(const char *elementPath, const char *data, int dataLen, void *callbackData)
{
	char buffer[1024];
	StringMap* myMapPtr = (StringMap*) callbackData;
	const char* sub;
	sub = strrchr((char*)elementPath, '/');
	if (sub == NULL)
	{
		sub = elementPath;
	}
	else
	{
		// skip "/"
		sub ++;
	}

	bool isWhiteSpace = true;
	for (int i = 0; i < dataLen; i ++)
	{
		if (data[i] != ' ' && data[i] != '\t' && data[i] != '\n')
		{
			isWhiteSpace = false;
		}
		buffer[i] = data[i];
	}

	buffer[dataLen] = '\0';
	if (!isWhiteSpace)
	{
		std::string myKey = sub;
		StringMap::iterator it = myMapPtr->find(std::string(sub));
		if (it != myMapPtr->end())
		{
			(*myMapPtr)[it->first] = buffer;
			FILE_LOG(logDEBUG)<<"[mySimpleXmlCallback]sub="<<sub<<";buffer="<<buffer;
		}
	}
	return 0;
}






/*
// testing...
void outputHeader()
{
    for (StringMap_Iterator it = stringMap.begin(); it != stringMap.end(); it ++)
    {
        cout<<"key:"<<it->first<<";value:"<<it->second<<endl;
    }
}



  bool MyS3::initializeIP()
    {
        char* host = "s3.amazonaws.com";
        //临时变量
        char  dns_buff[512];
        struct hostent  hostinfo, *phost;
        int   rc;
        int counter = 20;
        do
        {
            int result;
            if ((result=gethostbyname_r(host, &hostinfo, dns_buff, sizeof(dns_buff), &phost, &rc)) == 0)
            {
                inet_ntop(phost->h_addrtype, (u_int32_t*)(*phost->h_addr_list), ip, sizeof(ip));
                return true;
            }
            else
            {
                cout<<"\n failed\n"<<endl;
                switch (rc)
                {
                case HOST_NOT_FOUND:
                    cout<<"HOST_NOT_FOUND";
                    break;
                case NO_ADDRESS:
                    cout <<"NO_DATA or NO_ADDRESS";
                    break;
                case NO_RECOVERY:
                    cout<<"NO_RECOVERY";
                    break;
                case TRY_AGAIN:
                    cout<<"TRY_AGAIN";
                    break;
                default:
                    cout<<"unknown";
                }
                cout<<endl;
            }
            if (rc == TRY_AGAIN || rc == HOST_NOT_FOUND)
            {
                //unsigned int usec = rand()%1000000;
                //usleep(usec);
                //sleep(4);
                counter --;
            }
        }
        while (counter>0);
        return false;
    }
       std::string MyS3::urlEncode(const std::string& aContent)
    {
        std::string encoded;
        unsigned char c;
        unsigned char low, high;

        for (size_t i = 0; i < aContent.size(); i++)
        {
            c = aContent[i];
            //if (isalnum(c)||(c == '-' || c == '_' || c == '.' || c == '~' || c == '+'))
            if (isalnum(c)||(c == '-' || c == '_' || c == '.' || c == '~'))
            {
                encoded += c;
            }
            else
            {
                high = c / 16;
                low = c % 16;
                encoded += '%';
                encoded += (high < 10 ? '0' + high : 'A' + high - 10);
                encoded += (low < 10 ? '0' + low : 'A' + low - 10);
            }
        }
        return encoded;
    }

*/

